/* -*- mode: c; c-basic-offset: 4; -*- */
/* vi:set cindent ts=8 sts=4 sw=4: */
#include <assert.h>
#include <string.h>
#include <ctype.h>

#include "main.h"

#include "curs_main.h"
#include "keymap.h"
#include "lib.h"
#include "curs_main.h"

struct key_name_entry {
    int key; /* Special key code or ASCII value */
    char *name; /* Name of key */
};

struct key_name_entry key_name_table[] = {
    {' ',		"<Space>"},
    {CTRL('i'),		"<Tab>"},
    {KEY_ENTER,		"<Enter>"},
    {CTRL('j'),		"<Enter>"},
    {CTRL('m'),		"<CR>"},
    {ESC,		"M-"}, /* Should be followed by another key */
    {KEY_UP,		"<Up>"},
    {KEY_DOWN,		"<Down>"},
    {KEY_LEFT,		"<Left>"},
    {KEY_RIGHT,		"<Right>"},
    {KEY_BACKSPACE,    	"<BackSpace>"},
    {CTRL('h'),		"<BackSpace>"},
    {KEY_DC,            "<Delete>"},
    {KEY_HOME,		"<Home>"},
    {KEY_END,		"<End>"},
    {KEY_PPAGE, 	"<PageUp>"},
    {KEY_NPAGE, 	"<PageDown>"},

    {0,			NULL}
};

/*
 * Keymap for Relief.
 *
 * Except the entries for the line editor, the keymap entries are sorted by
 * the type of command (see the enumeration in keymap.h).
 * The command reference will display the keys for the first entry found if
 * the same command has more than one entry in the keymap.
 * Any key is allowed for `firstchar'.
 * Any key is allowed for `nextchar', except `ESC'.
 * Non-ASCII and non-graphical ASCII characters used _must_ have an entry
 * in `key_name_table', while graphical ASCII characters _can_ have an entry
 * (a character is graphical if isgrph(character) returns non-zero).
 * A command name's maximum length is `MAX_CMD_NAME_LEN'.
 */
struct keymap_entry keymap[] =
{
    {KEY_UP, 0, 0, ALL_MENUS, MOTION,
     "previous-entry", CMD_PREVIOUS_ENTRY,
     "up one entry"},
    {'k', 0, 0, ALL_MENUS, MOTION,
     "previous-entry", CMD_PREVIOUS_ENTRY,
     "up one entry"},
    {CTRL('p'), 0, 0, ALL_MENUS, MOTION,
     "previous-entry", CMD_PREVIOUS_ENTRY,
     "up one entry"},
    {'-', 0, 0, ALL_MENUS, MOTION,
     "previous-entry", CMD_PREVIOUS_ENTRY,
     "up one entry"},
    {KEY_DOWN, 0, 0, ALL_MENUS, MOTION,
     "next-entry", CMD_NEXT_ENTRY,
     "down one entry"},
    {'j', 0, 0, ALL_MENUS, MOTION,
     "next-entry", CMD_NEXT_ENTRY,
     "down one entry"},
    {CTRL('n'), 0, 0, ALL_MENUS, MOTION,
     "next-entry", CMD_NEXT_ENTRY,
     "down one entry"},
    {'+', 0, 0, ALL_MENUS, MOTION,
     "next-entry", CMD_NEXT_ENTRY,
     "down one entry"},
    {KEY_HOME, 0, 0, ALL_MENUS, MOTION,
     "first-entry", CMD_FIRST_ENTRY,
     "go to the first entry"},
    {'=', 0, 0, ALL_MENUS, MOTION,
     "first-entry", CMD_FIRST_ENTRY,
     "go to the first entry"},
    {KEY_END, 0, 0, ALL_MENUS, MOTION,
     "last-entry", CMD_LAST_ENTRY,
     "go to the last entry"},
    {'*', 0, 0, ALL_MENUS, MOTION,
     "last-entry", CMD_LAST_ENTRY,
     "go to the last entry"},
    {'H', 0, 0, ALL_MENUS, MOTION,
     "top-of-page", CMD_TOP_OF_PAGE,
     "go to the top entry of the page"},
    {'M', 0, 0, ALL_MENUS, MOTION,
     "middle-of-page", CMD_MIDDLE_OF_PAGE,
     "go to the middle entry of the page"},
    {'L', 0, 0, ALL_MENUS, MOTION,
     "bottom-of-page", CMD_BOTTOM_OF_PAGE,
     "go to the bottom entry of the page"},
    {'g', 0, 0, ALL_MENUS, MOTION,
     "goto", CMD_GO_TO,
     "enter the number of an entry to go to"},

    {ESC, 'k', 0, MENU_MAIN, SONG_MOVEMENT,
     "move-song-up", CMD_MOVE_SONG_UP,
     "move current song one entry upwards"},
    {ESC, KEY_UP, 0, MENU_MAIN, SONG_MOVEMENT,
     "move-song-up", CMD_MOVE_SONG_UP,
     "move current song one entry upwards"},
    {ESC, 'j', 0, MENU_MAIN, SONG_MOVEMENT,
     "move-song-down", CMD_MOVE_SONG_DOWN,
     "move current song one entry downwards"},
    {ESC, KEY_DOWN, 0, MENU_MAIN, SONG_MOVEMENT,
     "move-song-down", CMD_MOVE_SONG_DOWN,
     "move current song one entry downwards"},
    {ESC, '=', 0, MENU_MAIN, SONG_MOVEMENT,
     "move-song-first", CMD_MOVE_SONG_FIRST,
     "move current song to the top of the playlist"},
    {ESC, KEY_HOME, 0, MENU_MAIN, SONG_MOVEMENT,
     "move-song-first", CMD_MOVE_SONG_FIRST,
     "move current song to the top of the playlist"},
    {ESC, '*', 0, MENU_MAIN, SONG_MOVEMENT,
     "move-song-last", CMD_MOVE_SONG_LAST,
     "move current song to the bottom of the playlist"},
    {ESC, KEY_END, 0, MENU_MAIN, SONG_MOVEMENT,
     "move-song-last", CMD_MOVE_SONG_LAST,
     "move current song to the bottom of the playlist"},
    {ESC, 'H', 0, MENU_MAIN, SONG_MOVEMENT,
     "move-song-to-top-of-page", CMD_MOVE_SONG_TO_TOP_OF_PAGE,
     "move current song to the top of the page"},
    {ESC, 'M', 0, MENU_MAIN, SONG_MOVEMENT,
     "move-song-to-middle-of-page", CMD_MOVE_SONG_TO_MIDDLE_OF_PAGE,
     "move current song to the middle of the page"},
    {ESC, 'L', 0, MENU_MAIN, SONG_MOVEMENT,
     "move-song-to-bottom-of-page", CMD_MOVE_SONG_TO_BOTTOM_OF_PAGE,
     "move current song to the bottom of the page"},
    {ESC, 'g', 0, MENU_MAIN, SONG_MOVEMENT,
     "move-song-to", CMD_MOVE_SONG_TO,
     "enter an entry number to move the current song to"},

    {CTRL('y'), 0, 0, ALL_MENUS, SCROLLING,
     "scroll-up", CMD_SCROLL_UP,
     "scroll up one entry"},
    {CTRL('e'), 0, 0, ALL_MENUS, SCROLLING,
     "scroll-down", CMD_SCROLL_DOWN,
     "scroll down one entry"},
    {KEY_PPAGE, 0, 0, ALL_MENUS, SCROLLING,
     "page-up", CMD_PAGE_UP,
     "scroll one page upwards (backwards)"},
    {CTRL('b'), 0, 0, ALL_MENUS, SCROLLING,
     "page-up", CMD_PAGE_UP,
     "scroll one page upwards (backwards)"},
    {KEY_NPAGE, 0, 0, ALL_MENUS, SCROLLING,
     "page-down", CMD_PAGE_DOWN,
     "scroll one page downwards (forwards)"},
    {CTRL('f'), 0, 0, ALL_MENUS, SCROLLING,
     "page-down", CMD_PAGE_DOWN,
     "scroll one page downwards (forwards)"},
    {CTRL('u'), 0, 0, ALL_MENUS, SCROLLING,
     "half-page-up", CMD_HALF_PAGE_UP,
     "scroll a half page upwards (backwards)"},
    {CTRL('d'), 0, 0, ALL_MENUS, SCROLLING,
     "half-page-down", CMD_HALF_PAGE_DOWN,
     "scroll a half page downwards (forwards)"},

    {'?', 0, 0, MENU_MAIN | MENU_BROWSER, SEARCHING,
     "regexp-search-backward", CMD_REGEXP_SEARCH_BACKWARD,
     "regular expression search backwards"},
    {'/', 0, 0, MENU_MAIN | MENU_BROWSER, SEARCHING,
     "regexp-search-forward", CMD_REGEXP_SEARCH_FORWARD,
     "regular expression search forward"},
    {'N', 0, 0, MENU_MAIN | MENU_BROWSER, SEARCHING,
     "repeat-last-search-opposite", CMD_REPEAT_LAST_SEARCH_OPPOSITE,
     "repeat last search, in opposite direction"},
    {'n', 0, 0, MENU_MAIN | MENU_BROWSER, SEARCHING,
     "repeat-last-search", CMD_REPEAT_LAST_SEARCH,
     "repeat last search"},

    {'p', 0, 0, MENU_MAIN, PLAYING,
     "play-song", CMD_PLAY_SONG,
     "play the current song"},
    {CTRL('j'), 0, 0, MENU_MAIN, PLAYING,
     "play-song", CMD_PLAY_SONG,
     "play the current song"},
    {'P', 0, 0, MENU_MAIN, PLAYING,
     "play-toggled-songs", CMD_PLAY_TOGGLED_SONGS,
     "play the toggled songs"},

    {'A', 0, 0, MENU_MAIN, MISC,
     "add-directory", CMD_ADD_DIRECTORY,
     "add all songs in a directory to the playlist"},
    {'C', 0, 0, MENU_MAIN, MISC,
     "clear-playlist", CMD_CLEAR_PLAYLIST,
     "remove all songs from the playlist"},
    {'q', 0, 0, MENU_MAIN, MISC,
     "quit", CMD_QUIT,
     "exit the program"},
    {'b', 0, 0, MENU_MAIN, MISC,
     "append-playlist", CMD_APPEND_PLAYLIST,
     "append a playlist to the current"},
    {CTRL('l'), 0, 0, ALL_MENUS, MISC,
     "redraw-screen", CMD_REDRAW_SCREEN,
     "clear and redraw the screen"},
    {'t', 0, 0, MENU_MAIN, MISC,
     "toggle-song", CMD_TOGGLE_SONG,
     "toggle the current song"},
    {'c', 0, 0, MENU_MAIN, MISC,
     "toggle-continuous-flag", CMD_TOGGLE_CONTINUOUS_FLAG,
     "toggle the continuous flag"},
    {'r', 0, 0, MENU_MAIN, MISC,
     "toggle-random-flag", CMD_TOGGLE_RANDOM_FLAG,
     "toggle the random flag"},
    {'a', 0, 0, MENU_MAIN, MISC,
     "add-song", CMD_ADD_SONG,
     "add a song to the playlist"},
    {'d', 0, 0, MENU_MAIN, MISC,
     "remove-song", CMD_REMOVE_SONG,
     "remove the current song from the playlist"},
    {'h', 0, 0, MENU_MAIN | MENU_BROWSER | MENU_OPTIONS, MISC,
     "help", CMD_HELP,
     "get help for the current menu"},
    {'s', 0, 0, MENU_MAIN, MISC,
     "save-playlist", CMD_SAVE_PLAYLIST,
     "save the current playlist"},
    {'S', 0, 0, MENU_MAIN, MISC,
     "save-playlist-as", CMD_SAVE_PLAYLIST_AS,
     "save the current playlist in another file"},
    {'o', 0, 0, MENU_MAIN, MISC,
     "options-menu", CMD_OPTIONS,
     "enter the options menu"},
    {'D', 0, 0, MENU_MAIN | MENU_BROWSER, MISC,
     "change-directory", CMD_CHANGE_DIRECTORY,
     "change the current working directory"},
    {'w', 0, 0, MENU_MAIN, MISC,
     "write-song-names", CMD_WRITE_SONG_NAMES,
     "write the names of all songs in the current playlist to a file"},
    {'b', 0, 0, MENU_OPTIONS | MENU_HELP | MENU_BROWSER, MISC,
     "previous-menu", CMD_PREVIOUS_MENU,
     "return to the previous menu"},
    {CTRL('j'), 0, 0, MENU_OPTIONS, MISC,
     "edit-option", CMD_EDIT_OPTION,
     "edit non-boolean option"},
    {'r', 0, 0, MENU_OPTIONS, MISC,
     "reset-all-options", CMD_RESET_ALL_OPTIONS,
     "reset all options to their default values"},
    {'t', 0, 0, MENU_OPTIONS, MISC,
     "toggle-option", CMD_TOGGLE_OPTION,
     "toggle boolean option"},
    {CTRL('j'), 0, 0, MENU_BROWSER, MISC,
     "browse-directory", CMD_BROWSE_DIRECTORY,
     "browse the current directory"},
    {'v', 0, 0, MENU_MAIN | MENU_BROWSER, MISC,
     "sort", CMD_SORT,
     "sort the entries in increasing order"},
    {'V', 0, 0, MENU_MAIN | MENU_BROWSER, MISC,
     "sort-reverse", CMD_SORT_REVERSE,
     "sort the entries in decreasing order"},
    {'A', 0, 0, MENU_BROWSER, MISC,
     "add-current-directory", CMD_ADD_CURRENT_DIRECTORY,
     "add the current directory to the playlist"},
    {'a', 0, 0, MENU_BROWSER, MISC,
     "add-current-song", CMD_ADD_CURRENT_SONG,
     "add the current song to the playlist"},
    {'l', 0, 0, MENU_MAIN, MISC,
     "open-playlist", CMD_OPEN_PLAYLIST,
     "open playlist"},
    {'O', 0, 0, MENU_MAIN, MISC,
     "toggle-all-songs-off", CMD_TOGGLE_ALL_SONGS_OFF,
     "toggle all songs off"},
    {'Y', 0, 0, MENU_MAIN, MISC,
     "toggle-all-songs-on", CMD_TOGGLE_ALL_SONGS_ON,
     "toggle all songs on"},

    {CTRL('g'), 0, 0, LINE_EDITOR, MISC,
     "abort", CMD_ABORT,
     "abort"},
    {CTRL('a'), 0, 0, LINE_EDITOR, MISC,
     "beginning-of-line", CMD_BEGINNING_OF_LINE,
     "move to the start of the line"},
    {KEY_HOME, 0, 0, LINE_EDITOR, MISC,
     "beginning-of-line", CMD_BEGINNING_OF_LINE,
     "move to the start of the line"},
    {CTRL('e'), 0, 0, LINE_EDITOR, MISC,
     "end-of-line", CMD_END_OF_LINE,
     "move to the end of the line"},
    {KEY_END, 0, 0, LINE_EDITOR, MISC,
     "end-of-line", CMD_END_OF_LINE,
     "move to the end of the line"},
    {CTRL('f'), 0, 0, LINE_EDITOR, MISC,
     "forward-char", CMD_FORWARD_CHAR,
     "move forward a character"},
    {KEY_RIGHT, 0, 0, LINE_EDITOR, MISC,
     "forward-char", CMD_FORWARD_CHAR,
     "move forward a character"},
    {CTRL('b'), 0, 0, LINE_EDITOR, MISC,
     "backward-char", CMD_BACKWARD_CHAR,
     "move back a character"},
    {KEY_LEFT, 0, 0, LINE_EDITOR, MISC,
     "backward-char", CMD_BACKWARD_CHAR,
     "move back a character"},
    {CTRL('d'), 0, 0, LINE_EDITOR, MISC,
     "delete-char", CMD_DELETE_CHAR,
     "delete the character under the cursor"},
    {ESC, 'o', 0, LINE_EDITOR, MISC,
     "kill-whole-line", CMD_KILL_WHOLE_LINE,
     "kill the whole line"},
    {CTRL('k'), 0, 0, LINE_EDITOR, MISC,
     "kill-line", CMD_KILL_LINE,
     "kill from the cursor to the end of the line"},
    {ESC, CTRL('h'), 0, LINE_EDITOR, MISC,
     "backward-kill-word", CMD_BACKWARD_KILL_WORD,
     "kill the word behind the cursor"},
    {ESC, KEY_BACKSPACE, 0, LINE_EDITOR, MISC,
     "backward-kill-word", CMD_BACKWARD_KILL_WORD,
     "kill the word behind the cursor"},
    {ESC, 0x1f, 0, LINE_EDITOR, MISC,
     "backward-kill-word", CMD_BACKWARD_KILL_WORD,
     "kill the word behind the cursor"},
    {CTRL('h'), 0, 0, LINE_EDITOR, MISC,
     "backward-delete-char", CMD_BACKWARD_DELETE_CHAR,
     "delete the character behind the cursor"},
    {KEY_BACKSPACE, 0, 0, LINE_EDITOR, MISC,
     "backward-delete-char", CMD_BACKWARD_DELETE_CHAR,
     "delete the character behind the cursor"},
    {KEY_DC, 0, 0, LINE_EDITOR, MISC,
     "backward-delete-char", CMD_BACKWARD_DELETE_CHAR,
     "delete the character behind the cursor"},
    {CTRL('j'), 0, 0, LINE_EDITOR, MISC,
     "enter", CMD_ENTER,
     "enter"},
    {CTRL('x'), 0, 0, LINE_EDITOR, MISC,
     "browse", CMD_BROWSE,
     "browse the file system"},

    {0, 0, 0, 0, 0,
     NULL, KEYMAP_TERMINATOR,
     NULL}
};

struct cmd_ref_entry { 
    int cmd_id;
 /* The text to be displayed for the command, usually a shortened
    version of the command name */
    char *name_shown;
};

/* Pointer to an malloc()ed string containing the command reference for
   the main menu */
static char *menu_main_cmd_ref_line;
struct cmd_ref_entry menu_main_cmd_ref_entry[] = {
    {CMD_QUIT, "quit"},
    {CMD_HELP, "help"},
    {CMD_OPTIONS, "options"},
    {CMD_PLAY_SONG, "play"},
    {CMD_TOGGLE_SONG, "toggle"},
    {CMD_PLAY_TOGGLED_SONGS, "play toggled"},
    {CMD_ADD_DIRECTORY, "add dir"},
    {CMD_ADD_SONG, "add song"},
};

/* Pointer to an malloc()ed string containing the command reference for
   the options menu */
static char *menu_options_cmd_ref_line;
struct cmd_ref_entry menu_options_cmd_ref_entry[] = {
    {CMD_PREVIOUS_MENU, "back"},
    {CMD_EDIT_OPTION, "edit"},
    {CMD_TOGGLE_OPTION, "toggle"},
    {CMD_RESET_ALL_OPTIONS, "reset all"},
};

/* Pointer to an malloc()ed string containing the command reference for
   the help menu */
static char *menu_help_cmd_ref_line;
struct cmd_ref_entry menu_help_cmd_ref_entry[] = {
    {CMD_PREVIOUS_MENU, "back"},
};

/* Pointer to an malloc()ed string containing the command reference for
   the browser */
static char *menu_browser_cmd_ref_line;
struct cmd_ref_entry menu_browser_cmd_ref_entry[] = {
    {CMD_PREVIOUS_MENU, "back"},
    {CMD_HELP, "help"},
    {CMD_ADD_DIRECTORY, "add dir"},
    {CMD_ADD_SONG, "add song"},
    {CMD_CHANGE_DIRECTORY, "chdir"},
};

/*
 * Return a pointer to a buffer containing the command reference
 * for `menu'.
 */
const char *command_reference(int menu)
{
    int cmd_ref_line_size;
    char *cmd_ref_line;
    char buf[32];
    struct cmd_ref_entry *p = NULL;
    int entries = 0, i;

    /* For each menu, return the command reference if it's been initialized
       already, else make `p' point to the structure of cmd ref entries
       for the menu and `entries' contain the number of cmd ref entries. */
    if (menu == MENU_MAIN) {
	if (menu_main_cmd_ref_line != NULL)
	    return menu_main_cmd_ref_line;
	p = menu_main_cmd_ref_entry;
	entries = sizeof(menu_main_cmd_ref_entry)
	    / sizeof(struct cmd_ref_entry);
    } else if (menu == MENU_HELP) {
	if (menu_help_cmd_ref_line != NULL)
	    return menu_help_cmd_ref_line;
	p = menu_help_cmd_ref_entry;
	entries = sizeof(menu_help_cmd_ref_entry)
	    / sizeof(struct cmd_ref_entry);
    } else if (menu == MENU_BROWSER) {
	if (menu_browser_cmd_ref_line != NULL)
	    return menu_browser_cmd_ref_line;
	p = menu_browser_cmd_ref_entry;
	entries = sizeof(menu_browser_cmd_ref_entry)
	    / sizeof(struct cmd_ref_entry);
    } else if (menu == MENU_OPTIONS) {
	if (menu_options_cmd_ref_line != NULL)
	    return menu_options_cmd_ref_line;
	p = menu_options_cmd_ref_entry;
	entries = sizeof(menu_options_cmd_ref_entry)
	    / sizeof(struct cmd_ref_entry);
    } else
	assert(0);

    cmd_ref_line_size = COLS + 1;
    cmd_ref_line = safe_malloc(cmd_ref_line_size);
    for (i = 0, cmd_ref_line[0] = '\0'; i < entries; i++, p++) {
	get_key_descr_for_cmd(p->cmd_id, buf, sizeof(buf));
	if (strlen(cmd_ref_line) + strlen(buf) + strlen(":") +
	    strlen(p->name_shown) < cmd_ref_line_size) {
	    strcat(cmd_ref_line, buf);
	    strcat(cmd_ref_line, ":");
	    strcat(cmd_ref_line, p->name_shown);
	    /* Avoid always having a space in the last column, since one
	       extra character can cause the command reference to have 
	       one entry less (if there's more entries than it's room for) */
	    if (strlen(cmd_ref_line) + 1 < cmd_ref_line_size)
		strcat(cmd_ref_line, " ");
	}
    }

    if (menu == MENU_MAIN) {
	menu_main_cmd_ref_line = safe_strdup(cmd_ref_line);
	return menu_main_cmd_ref_line;
    } else if (menu == MENU_HELP) {
	menu_help_cmd_ref_line = safe_strdup(cmd_ref_line);
	return menu_help_cmd_ref_line;
    } else if (menu == MENU_BROWSER) {
	menu_browser_cmd_ref_line = safe_strdup(cmd_ref_line);
	return menu_browser_cmd_ref_line;
    } else if (menu == MENU_OPTIONS) {
	menu_options_cmd_ref_line = safe_strdup(cmd_ref_line);
	return menu_options_cmd_ref_line;
    } else
	assert(0);
    
    return NULL; /* not reached */
}

/*
 * Place a \0-terminated string describing the key(s) for `cmd_id' into `s'.
 * If there are more than one keymap entry for `cmd_id', the key(s) of the
 * first entry found will be used.
 */
void get_key_descr_for_cmd(int cmd_id, char *s, int slen)
{
    int i;

    /* Find the first keymap entry for `cmd_id' */
    for (i = 0; keymap[i].cmd_id != cmd_id &&
	     keymap[i].cmd_id != KEYMAP_TERMINATOR; i++)
	;
    assert(keymap[i].cmd_id != KEYMAP_TERMINATOR);
    key_description(i, s, slen);
}

/*
 * Place a \0-terminated string describing the key(s) for the `keymap_index'th
 * entry in `keymap' into `s'.
 */
void key_description(int keymap_index, char *s, int slen)
{ 
    char buf[32];
    bool found;
    int i = keymap_index;
    int j;

    for (j = 0, found = FALSE; key_name_table[j].key != 0; j++) {
	if (keymap[i].firstchar == key_name_table[j].key) {
	    snprintf(buf, sizeof(buf), key_name_table[j].name);
	    found = TRUE;
	    break;
	}
    }
    if (!found) {
	if (iscntrl(keymap[i].firstchar))
	    snprintf(buf, sizeof(buf), "C-%c", UNCTRL(keymap[i].firstchar));
	else
	    buf[0] = keymap[i].firstchar, buf[1] = '\0';
    }

    if (keymap[i].nextchar == 0) {
	snprintf(s, slen, buf);
	return;
    }

    for (j = 0, found = FALSE; key_name_table[j].key != 0; j++) {
	if (keymap[i].nextchar == key_name_table[j].key) {
	    snprintf(buf, sizeof(buf) - strlen(buf), key_name_table[j].name);
	    found = TRUE;
	    break;
	}
    }
    if (!found) {
	if (iscntrl(keymap[i].nextchar))
	    snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf),
		     "C-%c", UNCTRL(keymap[i].nextchar));
	else
	    snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf),
		     "%c", keymap[i].nextchar);
    }

    snprintf(s, slen, buf);
}
